{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TTFILE.DLL, routines for file transfer protocol}
unit FTLib;

interface
{$I teraterm.inc}

{$IFDEF Delphi}
uses Messages, WinTypes, WinProcs, TTTypes, TTFTypes, CommDlg, WinDos, Strings, DlgLib;
{$ELSE}
uses WinTypes, WinProcs, TTTypes, TTFTypes, CommDlg, Win31, WinDos, Strings, DlgLib;
{$ENDIF}

function GetTransFname(HWin: HWnd; FuncId: word; Option: Plongint;
                       FileVar: PFileVar; CurDir: PChar): Bool; export;
function GetNextFname(FileVar: PFileVar): bool;
function UpdateCRC(b: byte; CRC: word): word;
function UpdateCRC32(b: byte; CRC: longint): longint;
procedure FTLog1Byte(fv: PFileVar; b: byte);
procedure FTSetTimeOut(fv: PFileVar; T: integer);

implementation
const
  Id_FnDlg = 2100;
  Id_FnBox = 2101;
  Id_FnXCheck = 2102;
  Id_FnX1K = 2104;
  ID_FnBinary = 2105;
  ID_FnAppend = 2106;
  Id_FnDlg2 = 2200;
  Id_Fn2Binary = 2201;

{Hook function for file name dialog box}
function TFnHook(Dialog: HWnd; Message, WParam: Word; LParam: Longint): Bool; export;
var
  ofn: POpenFileName;
  Lo, Hi: word;
begin
  case Message of
    wm_InitDialog:
      begin
        ofn := POpenFileName(lParam);
        SetWindowLong(Dialog, DWL_USER, lParam);
        Lo := LoWord(ofn^.lCustData) and 1;
        Hi := HiWord(ofn^.lCustData);
        ShowDlgItem(Dialog,Id_FnBinary,Id_FnBinary);
        SetRB(Dialog,Lo,Id_FnBinary,Id_FnBinary);
        if Hi<>$FFFF then
        begin
          ShowDlgItem(Dialog,Id_FnAppend,Id_FnAppend);
          SetRB(Dialog,Hi and 1,Id_FnAppend,Id_FnAppend);
        end;
        TFnHook := true;
        exit;
      end;
    wm_Command:
      case wParam of
        id_Ok: begin
          ofn := POpenFileName(GetWindowLong(Dialog, DWL_USER));
          if ofn<>nil then
          begin
            GetRB(Dialog,Lo,Id_FnBinary,Id_FnBinary);
            Hi := HiWord(ofn^.lCustData);
            if Hi<>$FFFF then
              GetRB(Dialog,Hi,Id_FnAppend,Id_FnAppend);
            ofn^.lCustData := MakeLong(Lo,Hi);
          end;
        end;
        id_Cancel: ;
      end;
  end;
  TFnHook := False;
end;

function GetTransFname(HWin: HWnd; FuncId: word; Option: Plongint;
                       FileVar: PFileVar; CurDir: PChar): Bool;
var
  FNFilter: array [0..10] of char;
  FNameRec: TOpenFileName;
  TempDir: array[0..fsDirectory] of char;
  Ok: BOOL;
begin
  GetTransFname := false;
  {save current dir}
  GetCurDir(TempDir,0);

  FileVar^.FullName[0] := #0;
  FillChar(FNFilter, SizeOf(FNFilter), #0);  { Set up for double null at end }
  FillChar(FNameRec, SizeOf(TOpenFileName), #0);

  StrCopy(FileVar^.DlgCaption,'Tera Term: ');
  case FuncId of
    GTF_Send: StrCat(FileVar^.DlgCaption,'Send file');
    GTF_Log: StrCat(FileVar^.DlgCaption,'Log');
    GTF_BP: StrCat(FileVar^.DlgCaption,'B-Plus Send');
  else
    exit;
  end;

  StrCopy(FNFilter, 'all');
  StrCopy(@FNFilter[StrLen(FNFilter)+1], '*.*');      

  with FNameRec do
  begin
    lStructSize   := sizeof(TOpenFileName);
    hwndOwner     := HWin;
    lpstrFilter   := FNFilter;
    nFilterIndex := 1;
    lpstrFile := FileVar^.FullName;
    nMaxFile := SizeOf(FileVar^.FullName);
    lpstrInitialDir := CurDir;
    flags := ofn_ShowHelp or ofn_HideReadOnly or
             ofn_EnableTemplate;
    if FuncId<>GTF_BP then
    begin
      flags := flags or ofn_EnableHook;
      @lpfnHook := MakeProcInstance(@TFnHook, HInstance);
    end;
    FNameRec.lCustData := Option^;
    if FuncId<>GTF_Log then
    begin
      flags := flags or ofn_FileMustExist;
      FNameRec.lCustData := MakeLong(LoWord(Option^),$FFFF);
    end;
    lpstrTitle := FileVar^.DlgCaption;
    lpTemplateName := PChar(Id_FnDlg);
  end;
  FNameRec.hInstance := hinstance;

  Ok := GetOpenFileName(FNameRec);

  if Ok then
  begin
    if FuncId=GTF_Log then
      Option^ := FnameRec.lCustData
    else
      Option^ := MakeLong(LoWord(FnameRec.lCustData),HiWord(Option^));

    FileVar^.DirLen := FNameRec.nFileOffset;
    if CurDir<>nil then
    begin
      Move(FileVar^.FullName[0],CurDir[0],FileVar^.DirLen-1);
      CurDir[FileVar^.DirLen-1] := #0;
    end;
  end;

  {restore dir}
  SetCurDir(TempDir);

  GetTransFname := Ok;
end;


function GetNextFname(FileVar: PFileVar): bool;
var
  i: integer;
begin
with FileVar^ do begin
  {next file name exists?}
  GetNextFname := FNcount < NumFname;
  if FNcount >= NumFname then exit; {no more file name}

  inc(FNcount);
  if NumFname=1 then exit;

  i := DirLen;

  GlobalLock(FnStrMemHandle);
  while (PChar(FnStrMem)[FnPtr]<>#0) and
        (PChar(FnStrMem)[FnPtr]<>' ') do
  begin
    move(PChar(FnStrMem)[FnPtr],FullName[i],1);
    inc(FnPtr);
    inc(i);
  end;
  GlobalUnlock(FnStrMemHandle);

  inc(FnPtr);
  FullName[i] := #0;
end;
end;

function UpdateCRC(b: byte; CRC: word): word;
var
  i: integer;
begin
  CRC := CRC xor word(word(b) shl 8);
  for i := 1 to 8 do
    if (CRC and $8000)<>0 then
      CRC := (CRC shl 1) xor $1021
    else
      CRC := CRC shl 1;
  UpdateCRC := CRC;
end;

function UpdateCRC32(b: byte; CRC: longint): longint;
var
  i: integer;
begin
  CRC := CRC xor longint(b);
  for i := 1 to 8 do
    if (CRC and $00000001)<>0 then
      CRC := (CRC shr 1) xor $edb88320
    else
      CRC := CRC shr 1;
  UpdateCRC32 := CRC;
end;

procedure FTLog1Byte(fv: PFileVar; b: byte);
var
  d: array[0..2] of char;
begin
with fv^ do begin
  if LogCount= 16 then
  begin
    LogCount := 0;
    _lwrite(LogFile,#$0D#$0A,2);
  end;

  case b of
    $00..$9F: d[0] := char(b shr 4 + $30);
    $A0..$FF: d[0] := char(b shr 4 + $37);
  end;
  case b and $0F of
    $0..$9: d[1] := char(b and $0F + $30);
    $A..$F: d[1] := char(b and $0F + $37);
  end;
  d[2] := #$20;
  _lwrite(LogFile,d,3);
  inc(LogCount);
end;
end;

procedure FTSetTimeOut(fv: PFileVar; T: integer);
begin
  KillTimer(fv^.HMainWin, IdProtoTimer);
  if T=0 then exit;
  SetTimer(fv^.HMainWin, IdProtoTimer, T*1000, nil);
end;

end.