{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TTCMN.DLL, winsock interface
 Dynamic import version}

{Origin of this file is:

  Winsock.h file for Borland Pascal
  Conversion by Marc B. Manza
             Center for Applied Large-Scale Computing
  Requires Winsock.pas and Winsock.inc

  Send any comments/change requests/etc. to:
  manza@flash.poly.edu }

unit Winsock;

interface

uses WinTypes, WinProcs, WinsockH;

type
  Tclosesocket = function(s: SOCKET): integer;
  Tconnect = function(s: SOCKET; name: sockaddr; namelen: integer): integer;
  Thtonl = function(hostlong: u_long): u_long;
  Thtons = function(hostshort: u_short): u_short;
  Tinet_addr = function(cp: PChar): u_long;  
  Tioctlsocket = function(s: SOCKET; cmd: longint; argp: PLongint): integer;
  Trecv = function(s: SOCKET; buf: PChar; len, flags: integer): integer;
  Tselect = function(nfds: integer; readfds, writefds, exceptfds: Pfd_set; timeout: Ptimeval): longint;
  Tsend = function(s: SOCKET; buf: PChar; len, flags: integer): integer;
  Tsetsockopt = function(s: SOCKET; level, optname: integer; optval: PChar; optlen: integer): integer;
  T_socket = function(af, struct, protocol: integer): SOCKET;
  {Tgethostbyname = function(name : PChar): PHostEnt;}
  TWSAStartup = function(wVersionRequired: word; lpWSData: LPWSADATA): integer;
  TWSACleanup = function: integer;
  TWSAAsyncSelect = function(s: SOCKET; HWin: HWND; wMsg: word; lEvent: longint): integer;
  TWSAAsyncGetHostByName = function(HWin: HWND; wMsg: word; name, buf: PChar; buflen: integer): THandle;
  TWSACancelAsyncRequest = function(hAsyncTaskHandle: THandle): integer;
  TWSAGetLastError = function: integer;

var
  closesocket: Tclosesocket;
  connect: Tconnect;
  htonl: Thtonl;
  htons: Thtons;
  inet_addr: Tinet_addr;
  ioctlsocket: Tioctlsocket;
  recv: Trecv;
  select: Tselect;
  send: Tsend;
  setsockopt: Tsetsockopt;
  _socket: T_socket;
  {gethostbyname: Tgethostbyname;}
  WSAAsyncSelect: TWSAAsyncSelect;
  WSAAsyncGetHostByName: TWSAAsyncGetHostByName;
  WSACancelAsyncRequest: TWSACancelAsyncRequest;
  WSAGetLastError: TWSAGetLastError;
  WSAStartup: TWSAStartup;
  WSACleanup: TWSACleanup;

  HWinsock: THandle;
  WskUseCount: integer;

function WSAGetAsyncError(lParam: longint): word;
function WSAGetSelectError(lParam: longint): word;
procedure FD_SET(fd: socket; fdset: Pfd_set);
procedure FD_ZERO(fdset: Pfd_set);

function LoadWinsock: bool;
procedure CheckWinsock;
procedure FreeWinsock;

implementation

function WSAGetAsyncError(lParam: longint): word;
begin
  WSAGetAsyncError := HiWord(lParam);
end;

function WSAGetSelectError(lParam: longint): word;
begin
  WSAGetSelectError := HiWord(lParam);
end;

procedure FD_SET(fd: socket; fdset: Pfd_set);
begin
  if fdset^.fd_count < FD_SETSIZE then
  begin
    fdset^.fd_array[fdset^.fd_count] := fd;
    inc(fdset^.fd_count);
  end;
end;

procedure FD_ZERO(fdset: Pfd_set);
begin
  fdset^.fd_count := 0;
end;

const
  IdCLOSESOCKET           = 3;
  IdCONNECT               = 4;
  IdHTONL                 = 8;
  IdHTONS                 = 9;
  IdINET_ADDR             = 10;
  IdIOCTLSOCKET           = 12;
  IdRECV                  = 16;
  IdSELECT                = 18;
  IdSEND                  = 19;
  IdSETSOCKOPT            = 21;
  IdSOCKET                = 23;
  {IdGETHOSTBYNAME         = 52;}
  IdWSAASYNCSELECT        = 101;
  IdWSAASYNCGETHOSTBYNAME = 103;
  IdWSACANCELASYNCREQUEST = 108;
  IdWSAGETLASTERROR       = 111;
  IdWSASTARTUP            = 115;
  IdWSACLEANUP            = 116;

function LoadWinsock: bool;
var
  Err: boolean;
begin
  LoadWinsock := false;
  if HWinsock < 32 then
  begin
    WskUseCount := 0;

    HWinsock := LoadLibrary('WINSOCK.DLL');
    if HWinsock < 32 then exit;

    Err := false;

    @closesocket := GetProcAddress(HWinsock, PChar(IdCLOSESOCKET));
    if @closesocket=nil then Err := true;

    @connect := GetProcAddress(HWinsock, PChar(IdCONNECT));
    if @connect=nil then Err := true;

    @htonl := GetProcAddress(HWinsock, PChar(IdHTONL));
    if @htonl=nil then Err := true;

    @htons := GetProcAddress(HWinsock, PChar(IdHTONS));
    if @htons=nil then Err := true;

    @inet_addr := GetProcAddress(HWinsock, PChar(IdINET_ADDR));
    if @inet_addr=nil then Err := true;

    @ioctlsocket := GetProcAddress(HWinsock, PChar(IdIOCTLSOCKET));
    if @ioctlsocket=nil then Err := true;

    @recv := GetProcAddress(HWinsock, PChar(IdRECV));
    if @recv=nil then Err := true;

    @select := GetProcAddress(HWinsock, PChar(IdSELECT));
    if @select=nil then Err := true;

    @send := GetProcAddress(HWinsock, PChar(IdSEND));
    if @send=nil then Err := true;

    @setsockopt := GetProcAddress(HWinsock, PChar(IdSETSOCKOPT));
    if @setsockopt=nil then Err := true;

    @_socket := GetProcAddress(HWinsock, PChar(IdSOCKET));
    if @_socket=nil then Err := true;

    {@gethostbyname := GetProcAddress(HWinsock, PChar(IdGETHOSTBYNAME));
    if @gethostbyname=nil then Err := true;}

    @WSAAsyncSelect := GetProcAddress(HWinsock, PChar(IdWSAASYNCSELECT));
    if @WSAAsyncSelect=nil then Err := true;

    @WSAAsyncGetHostByName := GetProcAddress(HWinsock, PChar(IdWSAASYNCGETHOSTBYNAME));
    if @WSAAsyncGetHostByName=nil then Err := true;

    @WSACancelAsyncRequest := GetProcAddress(HWinsock, PChar(IdWSACANCELASYNCREQUEST));
    if @WSACancelAsyncRequest=nil then Err := true;

    @WSAGetLastError := GetProcAddress(HWinsock, PChar(IdWSAGETLASTERROR));
    if @WSAGetLastError=nil then Err := true;

    @WSAStartup := GetProcAddress(HWinsock, PChar(IdWSASTARTUP));
    if @WSAStartup=nil then Err := true;

    @WSACleanup := GetProcAddress(HWinsock, PChar(IdWSACLEANUP));
    if @WSACleanup=nil then Err := true;

    if Err then
    begin
      FreeLibrary(HWinsock);
      HWinsock := 0;
      exit;
    end;
  end;

  CheckWinsock;

  if HWinsock>=32 then
  begin
    inc(WskUseCount);
    LoadWinsock := true;
  end;
end;

procedure CheckWinsock;
var
  wVersionRequired : word;
  WSData : WSADATA;
begin
  if HWinsock< 32 then exit;
  wVersionRequired := 1*256+1;
  if (WSAStartup(wVersionRequired, @WSData)<>0) or
     (Lo(WSData.wVersion)<>1) or
     (Hi(WSData.wVersion)<>1) then
  begin
    WSACleanUp;
    FreeLibrary(HWinsock);
    HWinsock := 0;
  end;
end;

procedure FreeWinsock;
begin
  if WskUseCount<=0 then exit;

  WSACleanUp;
  dec(WskUseCount);

  if WskUseCount>0 then exit;
  if HWinsock >= 32 then
  begin
    FreeLibrary(HWinsock);
    HWinsock := 0;
  end;
end;

begin
  HWinsock := 0;
  WskUseCount := 0;
end.